var Assert = require('@kingjs/assert');

var Bootstrap = require('./bootstrap');
var Reflection = require('./reflection');
var Operators = require('./operators');
var Assign = require('./assign');
var Compose = require('./compose');
var Poset = require('./poset');

var flatten =  require('./flatten');
var pathReplace =  require('./pathReplace');

var defineFunctions = Bootstrap.defineFunctions;
var scorch = Bootstrap.scorch;
var pluck = Bootstrap.pluck;
var replaceEach = Bootstrap.replaceEach;
var toSet = Bootstrap.toSet;

var isFunction = Reflection.isFunction;
var isObjectLiteral = Reflection.isObjectLiteral;
var isArray = Reflection.isArray;
var isString = Reflection.isString;

var lambda1 = Operators.lambda1;
var returnArg0 = Operators.returnArg0;
var returnArg1 = Operators.returnArg1;

var assignDescriptorsDefaults = Assign.assignDescriptorsDefaults;
var assignDefaults = Assign.assignDefaults;

var composeDescriptors = Compose.composeDescriptors;

var flattenEncodedPoset = Poset.flattenEncodedPoset;
var walkEncodedPoset = Poset.walkEncodedPoset;
var walkPoset = Poset.walkPoset;

var mapDescriptorMetadataProperties = toSet('func', 'wrap', 'bases', 'defaults', 'depends', 'refs');
var mapDescriptorMetadataThunks = toSet('thunks');

defineFunctions(this, {  
  createMapDescriptorMetadata: function(/* ... */) {

    // normalize first argument
    if (isFunction(arguments[0]))
      arguments[0] = { func: arguments[0] };
    
    var result = { };
    assignDescriptorsDefaults(mapDescriptorMetadataProperties, result, arguments);
    composeDescriptors(mapDescriptorMetadataThunks, result, arguments);
    return result;
  },
  
  // mapDescriptor(action, [ ... [ { name0: v0, name1: v1, ..., nameM: vM } ] ... ]) =>
  //   action.call(this, 'name0', v0)
  //   action.call(this, 'name1', v1)
  //   ...
  //   action.call(this, 'nameM', vM)
  // defaults:
  //   A descriptor with name '$' contains default descriptor values. If any of it's properties 
  //   start with '$', then they are evaluated with (name, descriptor) to generate the value
  //   except if name is '$' in which case non-object-literals are wrapped with name being the value of '$'
  mapDescriptors: function(actions, descriptorPile) {
    Assert.that(arguments.length == 2);

    if (!descriptorPile)
      return;
        
    // for each action pass...
    var result;
    flatten(function(action) {    
      if (isFunction(action) || !action)
        action = { func: action };

      if (!action.func)
        action.func = returnArg1;
      
      result = { };
      var dependencies = { };
      
      // for each encoded descriptor; e.g. name$ex0$ex1
      flatten(function(encodedDescriptors) {
        
        // pluck meta descriptor properties; inherit defaults from action
        var meta = exports.createMapDescriptorMetadata(scorch({
          func: pluck(encodedDescriptors, '$func'),
          wrap: pluck(encodedDescriptors, '$wrap'),
          bases: pluck(encodedDescriptors, '$bases'),
          defaults: pluck(encodedDescriptors, '$defaults'),
          depends: pluck(encodedDescriptors, '$depends') || { },
          thunks: pluck(encodedDescriptors, '$thunks') || { }      
        }), action);
        
        meta.bases = flattenEncodedPoset(meta.bases);
               
        walkEncodedPoset(function(descriptors, name, baseNames) {        
          var descriptor = descriptors[name];
          
          // flatten
          result[name] = descriptor;
          
          // get (and normalize) descriptor
          if (!isObjectLiteral(descriptor)) {
            if (!meta.wrap)
              return;
            
            // normalize descriptor
            var normalizedDescriptor = { };
            if (isFunction(meta.wrap))
              normalizedDescriptor = meta.wrap.call(this, descriptor);
            else
              normalizedDescriptor[meta.wrap] = descriptor;
            
            result[name] = descriptor = normalizedDescriptor;
          }
          
          // propogate meta for plucking in second pass
          descriptor.$ = meta;
          
          // assign base values
          baseNames.forEach(function(baseName) {
            assignDefaults(descriptor, meta.bases[baseName])
          });
          
          // assign default values
          assignDefaults(descriptor, meta.defaults);
          
          // gather dependencies
          flatten(function(dependencyName) {
            var refs = descriptor[dependencyName];
            if (!refs)
              return;
            
            // mark for denormalizing (e.g. base vs implements)
            if (!isArray(refs)) {
              refs = [ refs ];
              refs.denormalize = true;
            }
            
            dependencies[name] = refs;
            descriptor[dependencyName] = refs;
          }, Object.keys(meta.depends))
          
        }, encodedDescriptors)
      }, descriptorPile);
            
      // depth first walk of decoded descriptor poset...
      walkPoset(function(name, baseNames) {
        var descriptor = result[name];      
        
        if (isObjectLiteral(descriptor)) {
          
          var meta = pluck(descriptor, '$');
          
          // resolve dependencies
          flatten(function(dependencyName) {
            var refs = descriptor[dependencyName];
            if (!refs)
              return;
            
            // allow projection of resolved descriptor
            var dependencySelector = meta.depends[dependencyName] || returnArg0;
            replaceEach(function(ref) {
              if (!isString(ref))
                return ref;
              
              var dependency = result[ref];
              Assert.that(dependency);
              return dependencySelector(dependency);
            }, refs);
            
            // denormalize
            if (refs.denormalize)
              descriptor[dependencyName] = refs[0];            
          }, Object.keys(meta.depends));
                    
          // initialize values
          for (var key in descriptor) {
            var value = descriptor[key];
            
            // procedural/contextual defaults
            if (isFunction(value) && value.name == '$')        
              value = value.call(this, name);
            
            // thunk
            var thunk = meta.thunks[key];
            if (thunk)
              value = thunk.call(this, name, key, value);
            
            // delete undefined values
            if (value === undefined) {
              delete descriptor[key];
              continue;
            }
            
            descriptor[key] = value;
          }
        }
        
        result[name] = action.func.call(this, name, descriptor);
      }, result, dependencies)
      
      // resolve/select references
      var refs = action.refs;
      if (refs) {
        for (var encodedPath in refs) {
          for (var name in result) {
            var descriptor = result[name];
            var projection = refs[encodedPath];
            
            var path = encodedPath.split('.');
            pathReplace(function(ref) {
              if (!isString(ref))
                return ref;
              return projection(result[ref]);
            }, descriptor, path);
          }
        }
      }
          
      // output of this action becomes input of next action
      return descriptorPile = result;
      
    }, [ actions ]);
    
    return result;
  },
});